﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Diagnostics;

namespace Reversi
{
    // 着手データクラス
    struct Move
    {
        // 着手のパス
        public static Move PASS = new Move(-1, -1);

        // 着手の位置
        public int x;
        public int y;

        // コンストラクタ(位置を指定)
        public Move(int x, int y)
        {
            this.x = x;
            this.y = y;
        }

        public bool equal(Move move)
        {
            return ((this.x == move.x) && (this.y == move.y)) ? true : false;
        }
    }

    // ひっくり返しフラグクラス
    struct ReverseFlag
    {
    	// 着手した石
        public const int MOVE_STONE = 1;
        // ひっくり返った石
        public const int REVERSED_STONE = 2;
        
        // 着手した石フラグ (64ビット整数)
        public ulong moveStone;
        // ひっくり返した石フラグ (64ビット整数)
        public ulong reversedStone;
        
        // リセット
        public void reset()
        {
            moveStone = 0;
            reversedStone = 0;
        }
        // 指定された場所の石を取得する
        public int getStone(int x, int y)
        {
            ulong mask = 1ul << (y * 8 + x);
            if ((moveStone & mask) != 0) return MOVE_STONE;
            if ((reversedStone & mask) != 0) return REVERSED_STONE;
            return 0;
        }

        // 指定された場所の石を設定する
        public void setStone(int x, int y, int attribute)
        {
            ulong mask = 1ul << (y * 8 + x);

            if (attribute == MOVE_STONE)
            {
                moveStone |= mask;
            }
            else if (attribute == REVERSED_STONE)
            {
                reversedStone |= mask;
            }
        }
    }

    // 盤面クラス
    struct Board
    {
        // 先手
        public const int FIRST_MOVE = 1;
        // 後手
        public const int SECOND_MOVE = -1;
        // 石が無い
        public const int NO_STONE = 0;
        // 引き分け
        public const int DRAW = 0;
        // 路の数 (通常の8路盤)
        public const int SIZE = 8;

        // 先手の石フラグ(64ビット整数)
        public ulong firstStone;
        // 後手の石フラグ(64ビット整数)
        public ulong secondStone;
        // 対局開始からの手順数
        public int turn;

        // 手番持ちのプレイヤー（先手 or 後手）
        public int turnHolder
        {
            get
            {
                return ((turn % 2) == 0) ? FIRST_MOVE : SECOND_MOVE;
            }
        }
        // 指定された場所の石を取得する
        public int getStone(int x, int y)
        {
            ulong mask = 1ul << (y * 8 + x);

            if ((firstStone & mask) != 0) return FIRST_MOVE;
            if ((secondStone & mask) != 0) return SECOND_MOVE;
            return NO_STONE;
        }

        // 指定された場所の石を設定する
        public void setStone(int x, int y, int owner)
        {
            ulong mask = 1ul << (y * 8 + x);

            if (owner == FIRST_MOVE)
            {
                firstStone |= mask;
                secondStone &= ~mask;
            }
            else if (owner == SECOND_MOVE)
            {
                firstStone &= ~mask;
                secondStone |= mask;
            }
        }

        // 勝者
        public int winner
        {
            get
            {
                int f = countStone(FIRST_MOVE);
                int s = countStone(SECOND_MOVE);

                if (f > s)
                {
                    return FIRST_MOVE;
                }
                else if(f < s)
                {
                    return SECOND_MOVE;
                }
                else
                {
                    return DRAW;
                }
            }
        }

        // 盤面の範囲内か判定する
        private static bool isInBoard(int x, int y)
        {
            if ((x < 0) || (x >= SIZE)) return false;
            if ((y < 0) || (y >= SIZE)) return false;
            return true;
        }

        // 盤面を初期状態にリセットする
        public void reset()
        {
            firstStone = 0;
            secondStone = 0;
            setStone(SIZE / 2 - 1, SIZE / 2 - 1, SECOND_MOVE);
            setStone(SIZE / 2 - 1, SIZE / 2    , FIRST_MOVE);
            setStone(SIZE / 2    , SIZE / 2 - 1, FIRST_MOVE);
            setStone(SIZE / 2    , SIZE / 2    , SECOND_MOVE);

            turn = 0;
        }
        
        // 着手可能な手があるか？
        public bool hasAvailableMove()
        {
            ulong nextMoves = enumNextMoves();

            return (nextMoves != 0);
        }

        // 有効な着手か判定する
        public bool isAvailableMove(Move move)
        {
            int X = move.x;
            int Y = move.y;

            // 彼我のどちらが先手/後手か
            int mine = turnHolder;
            int his = (mine == FIRST_MOVE) ? SECOND_MOVE : FIRST_MOVE;

            // 範囲内で石の無い位置であることをチェック
            if (!isInBoard(X, Y)) return false;
            if (getStone(X, Y) != NO_STONE) return false;

            // 八方の石をチェック
            for (int dx = -1; dx <= 1; dx++)
            {
                for (int dy = -1; dy <= 1; dy++)
                {
                    if ((dx == 0) && (dy == 0)) continue;

                    for (int i = 1; i < SIZE; i++)
                    {
                        int x = X + i * dx;
                        int y = Y + i * dy;
                        if (!isInBoard(x, y)) break;
                        int s = getStone(x, y);
                        if (s == NO_STONE) break;
                        else if (s == his) continue;
                        else if (s == mine)
                        {
                            if (i == 1) break;
                            return true;
                        }
                        else break; // abnormal case
                    }
                }
            }
            return false;
        }

        // 着手する
        public void doMove(Move move)
        {
            int X = move.x;
            int Y = move.y;

            // 彼我のどちらが先手/後手か
            int mine = turnHolder;
            int his = (mine == FIRST_MOVE) ? SECOND_MOVE : FIRST_MOVE;
            // 手順数を進める
            turn++;

            // 範囲内で石の無い位置であることをチェック
            if (!isInBoard(X, Y)) return;
            if (getStone(X, Y) != NO_STONE) return;

            // 着手
            setStone(X, Y, mine);

            // 八方の石をチェック
            for (int dx = -1; dx <= 1; dx++)
            {
                for (int dy = -1; dy <= 1; dy++)
                {
                    if ((dx == 0) && (dy == 0)) continue;

                    for (int i = 1; i < SIZE; i++)
                    {
                        int x = X + i * dx;
                        int y = Y + i * dy;
                        if (!isInBoard(x, y)) break;
                        int s = getStone(x, y);
                        if (s == NO_STONE) break;
                        else if (s == his) continue;
                        else if (s == mine)
                        {
                            // ひっくり返す
                            for (int j = 1; j < i; j++)
                            {
                                setStone(X + j * dx, Y + j * dy, mine);
                            }
                            break;
                        }
                        else break; // abnormal case
                    }
                }
            }
            return;
        }

        // 着手する (ひっくり返しフラグを返す)
        public void doMove(Move move, ref ReverseFlag reserse)
        {
            reserse.reset();
            Board old = this;

            doMove(move);

            for (int x = 0; x < SIZE; x++)
            {
                for (int y = 0; y < SIZE; y++)
                {
                    if (old.getStone(x, y) != this.getStone(x, y))
                    {
                        reserse.setStone(x, y, ReverseFlag.REVERSED_STONE);
                    }
                }
            }
            reserse.setStone(move.x, move.y, ReverseFlag.MOVE_STONE);
        }

        // 石の数を数える
        public int countStone(int player)
        {
            int count = 0;

            for (int x = 0; x < SIZE; x++)
            {
                for (int y = 0; y < SIZE; y++)
                {
                    if(getStone(x,y) == player) count++;
                }
            }
            return count;
        }

        // 対局終了判定
        public bool hasFinished()
        {
            // 盤が石で埋まれば終了
            if (countStone(NO_STONE) == 0)
            {
                return true;
            }
            // 片方の石がゼロになれば終了
            if(countStone(FIRST_MOVE) == 0 || countStone(SECOND_MOVE) == 0)
            {
                return true;
            }
            // 双方着手不可能な場合
            if (!hasAvailableMove())
            {
                Board nextBoard = this;
                nextBoard.doMove(Move.PASS);
                if(!nextBoard.hasAvailableMove())
                {
                    return true;
                }
            }
            return false;
        }

        // 次の局面を列挙する
        public ulong enumNextMoves()
        {
            ulong nextMoves = 0;

            // 盤面から有効な着手を探す
            for (int x = 0; x < SIZE; x++)
            {
                for (int y = 0; y < SIZE; y++)
                {
                    Move move = new Move(x, y);
                    if (isAvailableMove(move))
                    {
                        // 有効な着手を記憶
                        nextMoves |= (1ul << (y * 8 + x));
                    }
                }
            }
            return nextMoves;
        }
    }
}
